
# q-expansion of newform 640.2.q.d, downloaded from the LMFDB on 02 June 2024.

# We generate the q-expansion using the Hecke eigenvalues a_p at the primes.
# Each a_p is given as a linear combination
# of the following basis for the coefficient ring.

def make_data():

    from sage.all import prod, floor, prime_powers, gcd, QQ, primes_first_n, next_prime, RR

    def discrete_log(elts, gens, mod):
        # algorithm 2.2, page 16 of https://arxiv.org/abs/0903.2785
        def table_gens(gens, mod):
            T = [1]
            n = len(gens)
            r = [None]*n
            s = [None]*n
            for i in range(n):
                beta = gens[i]
                r[i] = 1
                N = len(T)
                while beta not in T:
                    for Tj in T[:N]:
                        T.append((beta*Tj) % mod)
                    beta = (beta*gens[i]) % mod
                    r[i] += 1
                s[i] = T.index(beta)
            return T, r, s
        T, r, s = table_gens(gens, mod)
        n = len(gens)
        N = [ prod(r[:j]) for j in range(n) ]
        Z = lambda s: [ (floor(s/N[j]) % r[j]) for j in range(n)]
        return [Z(T.index(elt % mod)) for elt in elts]
    def extend_multiplicatively(an):
        for pp in prime_powers(len(an)-1):
            for k in range(1, (len(an) - 1)//pp + 1):
                if gcd(k, pp) == 1:
                    an[pp*k] = an[pp]*an[k]
    from sage.all import PolynomialRing, NumberField
    R = PolynomialRing(QQ, "x")
    f = R(poly_data)
    K = NumberField(f, "a")
    betas = [K.gens()[0]**i for i in range(len(poly_data))]
    convert_elt_to_field = lambda elt: sum(c*beta for c, beta in zip(elt, betas))
    # convert aps to K elements
    primes = primes_first_n(len(aps_data))
    good_primes = [p for p in primes if not p.divides(level)]
    aps = map(convert_elt_to_field, aps_data)
    if not hecke_ring_character_values:
        # trivial character
        char_values = dict(zip(good_primes, [1]*len(good_primes)))
    else:
        gens = [elt[0] for elt in hecke_ring_character_values]
        gens_values = [convert_elt_to_field(elt[1]) for elt in hecke_ring_character_values]
        char_values = dict([(
            p,prod(g**k for g, k in zip(gens_values, elt)))
            for p, elt in zip(good_primes, discrete_log(good_primes, gens, level))
            ])
    an_list_bound = next_prime(primes[-1])
    an = [0]*an_list_bound
    an[1] = 1
    
    from sage.all import PowerSeriesRing
    PS = PowerSeriesRing(K, "q")
    for p, ap in zip(primes, aps):
        if p.divides(level):
            euler_factor = [1, -ap]
        else:
            euler_factor = [1, -ap, p**(weight - 1) * char_values[p]]
        k = RR(an_list_bound).log(p).floor() + 1
        foo = (1/PS(euler_factor)).padded_list(k)
        for i in range(1, k):
            an[p**i] = foo[i]
    extend_multiplicatively(an)
    return PS(an)
level = 640
weight = 2
poly_data = [1, 0, 1]

# The basis for the coefficient ring is just the power basis
# in the root of the defining polynomial above.
hecke_ring_character_values = [[511, [1, 0]], [261, [0, 1]], [257, [-1, 0]]]
aps_data = [[0, 0], [1, 1], [-1, -2], [0, 0], [-3, -3], [-3, -3], [0, -4], [-1, 1], [8, 0], [-3, 3], [0, 0], [-3, 3], [0, 0], [3, -3], [0, 2], [9, -9], [9, 9], [5, -5], [-3, -3], [0, -6], [6, 0], [-8, 0], [9, 9], [0, -12], [0, 12], [-3, -3], [0, 0], [-9, 9], [1, -1], [0, -8], [0, -6], [-9, 9], [2, 0], [-7, -7], [-3, -3], [0, 18], [9, 9], [9, 9], [-8, 0], [9, 9], [3, -3], [1, 1], [24, 0], [0, -12], [5, -5], [0, 2], [11, -11], [0, -6], [9, 9], [-7, -7], [22, 0], [-24, 0], [-18, 0], [9, 9], [0, -8], [16, 0], [9, -9], [-16, 0], [-3, 3], [0, -12], [15, -15], [9, -9], [-3, -3], [0, -6], [-6, 0], [-7, -7], [5, 5], [0, -24], [19, -19], [5, -5], [0, -16], [0, 18], [0, 18], [-3, 3], [1, 1], [0, 10], [-15, -15], [9, 9], [30, 0], [0, 0], [15, -15], [5, 5], [24, 0], [0, 36], [0, 10], [-9, 9], [-18, 0], [18, 0], [-3, 3], [0, -30], [5, 5], [24, 0], [-24, 0], [-15, -15], [-29, 29], [-32, 0], [9, -9], [0, -24], [-9, 9], [1, -1], [-3, -3], [9, 9], [-19, -19], [0, 24], [-11, -11], [0, 24], [-9, 9], [0, 32], [0, -30], [0, -36], [0, 42], [-27, 27], [-10, 0], [13, 13], [0, 2], [30, 0], [-27, -27], [32, 0], [9, 9], [-21, 21], [29, 29], [0, 12], [9, -9], [-13, 13], [-5, 5], [-3, 3], [13, 13], [-24, 0], [-24, 0], [-3, -3], [19, -19], [8, 0], [0, 0], [33, -33], [0, 48], [-18, 0], [-23, 23], [33, 33], [-19, -19], [0, 0], [37, 37], [-39, -39], [24, 0], [31, -31], [-35, 35], [0, 42], [-15, 15], [-38, 0], [-7, -7], [0, -22], [-3, -3], [-6, 0], [21, 21], [8, 0], [-21, 21], [48, 0], [0, -54], [30, 0], [54, 0], [-27, 27], [-27, -27], [-22, 0], [48, 0], [9, 9], [0, -28], [-16, 0], [16, 0], [9, -9]]
